/*
  The TrackStore class contains all the methods necessary to prepare 
  the tracks to feed the alignment algorithm

  Methods are:

  PREPARESTORE

  GETTRACKINFO


*/

#include "TrackStore.h"
#include "Track.h"
#include "Line.h"
#include "constants.h"

using namespace std;

TrackStore::TrackStore()
{  
  cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << endl; 
  cout << "%                                     %" << endl;
  cout << "%    Entering into the TrackStore     %" << endl;
  cout << "%                                     %" << endl;
  cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << endl; 
  cout << "" << endl;
}
TrackStore::~TrackStore()
{
}


//=================================================================
// PrepareStore
//=================================================================
//
// Create a set of toy tracks according to the detector geometry
// and to the number of requested events
//
//=================================================================

void TrackStore::PrepareStore(Tracks* my_tracks, const Planes& geom, int n_event)
{
  int event = 0;

  while (event<n_event)  
  {
    if (debug_mode) cout << "Generate event number " << event << endl; 

    bool sc = TrackStore::GenerEvent(my_tracks,geom,event);
    if (sc) event++;
  }

  if (debug_mode) cout << "Number of tracks generated " << my_tracks->size() << endl; 
}



//=================================================================
// GenerEvent
//=================================================================
//
// Create a set of toy tracks for one particular event, all the 
// tracks are supposed to come from the same primary vertex
// (...This could be upgraded...)
//
//=================================================================

bool TrackStore::GenerEvent(Tracks* my_event_tracks,const Planes& geom, int n_event)
{

  int    m_ntracks = 0, ntracks = 0;
  double prim_ver[3];

  TF1* m_GaussGenerator = new TF1("m_gaussGenerator","gaus",-30.,30.);

// Number of tracks in the event

  m_GaussGenerator->SetParameters(1,0,20);

  while (m_ntracks < 5) m_ntracks = int(m_GaussGenerator->GetRandom());

  if (debug_mode) cout << "Event with " << m_ntracks << " tracks" << endl;

// Generate primary vertex
  
  m_GaussGenerator->SetParameters(1,0,1);
  
  prim_ver[0] = m_GaussGenerator->GetRandom()*0.01;
  prim_ver[1] = m_GaussGenerator->GetRandom()*0.01;
  prim_ver[2] = m_GaussGenerator->GetRandom()*50.0-dist_PV;  

    
  if (debug_mode) {
    
    cout << "Primary vertex coordinates (in mm): " << endl;
    cout << "x_v = " << prim_ver[0] << endl;
    cout << "y_v = " << prim_ver[1] << endl;
    cout << "z_v = " << prim_ver[2] << endl;
  }
  
// Now generate the tracks

  while (ntracks <m_ntracks) {

    double x2 = ((float(rand()%10000))/10000. - 0.5)*planeX;
    double y2 = ((float(rand()%10000))/10000. - 0.5)*planeY;

    Line xLine(prim_ver[0],x2,prim_ver[2], geom.back().z());
    Line yLine(prim_ver[1],y2,prim_ver[2], geom.back().z());

    //    double side = m_GaussGenerator->GetRandom();
    bool isForward;

    //    (side <= 0.) ? isForward = true : isForward = false;

    isForward = true;

    Track* new_track = new Track(xLine,yLine);

    new_track->setOrigin(prim_ver);
    new_track->setDirection(isForward);

    // Enough hits on the track ??

    bool sc  = TrackStore::CheckTrack(new_track,geom);

    if (sc)
    {
      new_track->setEventnum(n_event);
      new_track->setTracknum(ntracks);

      if (debug_mode) cout << "Track evt num." << new_track->getEventnum() << endl; 

      my_event_tracks->push_back(new_track);

      ntracks++;
    }
  }

  return true;
}




//=================================================================
// CheckTrack
//=================================================================
//
// Control that there is a sufficient number of hits in the 
// active area (set by default to 3 hits)
//
//=================================================================


bool TrackStore::CheckTrack(Track* intrack, const Planes& geom)
{

  int nStation = geom.size();
  int nhit = 0, nhit_total = 0;

  bool accepted_track = false;

  double track_orig[3];

  intrack->getOrigin(&track_orig[0]);

  double v_z = track_orig[2];

  bool   track_direction = intrack->getDirection();

  //  cout << endl;

  for (int iStation = 0; iStation < nStation; ++iStation)
  { 
    // Check if the track point is inside the sensor active area

    nhit_total++;

    double zplane = geom[iStation].z();

    //    cout << "X = " << intrack->x(zplane) << " / Y = " << intrack->y(zplane) 
    //	 << " / Z = " << zplane << endl;
    
    if ((zplane < v_z) && (track_direction == true)) continue;
    if ((zplane > v_z) && (track_direction == false)) continue;
    if (!(geom[iStation].isInside(intrack->x(zplane),intrack->y(zplane)))) continue;

    nhit++;
  }

  if (nhit >= 3) accepted_track = true;

  if (debug_mode) cout << nhit_total << " / " << nhit << endl;

  return accepted_track;
}

