
/** @ myAlg.cpp 
 *
 *  The main alignment code
 *  
 *  Interface with Millepede is made by MilleTool
 *
 *  @author M.Needham
 */

#include <iostream>
#include <iomanip>

// For ROOT

#include "TFile.h"
#include "TH1F.h"
#include "TH2F.h"
#include "TF1.h"

// Knossos includes

#include "myAlg.h"
#include "MilleTool.h"
#include "Track.h"
#include "TrackStore.h"
#include "constants.h"
#include "fun.h"


using namespace std;

myAlg::myAlg():
 m_rootFile(0)
{
  // constructer
}

myAlg::~myAlg()
{
 // destructer
  delete m_GaussGenerator;
  delete m_align;
  delete my_tracks;
  delete input_tracks;
}

/*
  INITIALIZE:  Initialization of the ROOTfile and the detector geometry
*/

bool myAlg::initialize(const jobParams& tparams)
{
  m_params = tparams;

  // create root file

  m_rootFile = new TFile(params().outputFile().c_str(),
                         "RECREATE","my ROOT file");
  m_rootFile->SetCompressionLevel(1);

  m_GaussGenerator = new TF1("m_gaussGenerator","gaus",-10.,10.);
  m_GaussGenerator->SetParameters(1,0,1);

  // here we generate the ideal geometry, using the options 
  // set in constants.h

  for (unsigned int iStation = 0; iStation < nStation; ++iStation)
  {   
    double z = firstZ + iStation*dzStation;   //station
    Plane newPlane(-planeX/2,planeX/2,-planeY/2,planeY/2,z); 
    m_idealGeometry.push_back(newPlane);  // The detector plane is added to the plane
  }


  // PrepareStore fills the Store according to TrackCuts

  my_tracks    = new TrackStore();
  input_tracks = new Tracks();

  // make a tool to access millepede
  m_align  = new MilleTool(); 

  // book some histos...
  m_txHisto  = new TH1F("tx histo","tx histo",100,-1.,1.);
  m_xyHisto = new TH2F("x vs y", "x vs y", 100, -planeX, 
                       planeX, 100, -planeY, planeY);


  double misal_scales[6];

  (offsetX == 0) ? misal_scales[0] = 1. : misal_scales[0] = 4*offsetX;
  (offsetY == 0) ? misal_scales[1] = 1. : misal_scales[1] = 4*offsetY;
  (offsetZ == 0) ? misal_scales[2] = 1. : misal_scales[2] = 4*offsetZ;
  (offsetA == 0) ? misal_scales[3] = 1. : misal_scales[3] = 4*offsetA;
  (offsetB == 0) ? misal_scales[4] = 1. : misal_scales[4] = 4*offsetB;
  (offsetC == 0) ? misal_scales[5] = 1. : misal_scales[5] = 4*offsetC;

 for (unsigned int histID = 0; histID < 6; ++histID)
 {
    char histoName[100];
    char histoName2[100];
    char histoName3[100];

    sprintf(histoName, "mis_before_%d",histID);
    TH1F* aHisto = new TH1F(histoName, histoName, 100, -misal_scales[histID], misal_scales[histID]);
    m_mis_bef.push_back(aHisto);

    sprintf(histoName2, "mis_after_%d",histID);
    TH1F* aHisto2 = new TH1F(histoName2, histoName2, 100, -0.4*misal_scales[histID], 0.4*misal_scales[histID]);
    m_mis_aft.push_back(aHisto2);
                                                                                
    sprintf(histoName3, "pullHisto%d",histID);
    TH1F* aHisto3 = new TH1F(histoName3, histoName3, 40, -5, 5);
    m_pullHistos.push_back(aHisto3);
  }

  return true;	  
}


/*
  EXECUTE:  
*/

bool myAlg::execute()
{
  
  for (unsigned int iRun = 0;  iRun < m_params.nRun(); ++iRun)
  {

    //    
    // First of all the misaligned geometry is generated (here only x offsets )
    //

    m_offsetX.resize(nStation);
    m_offsetY.resize(nStation);
    m_offsetZ.resize(nStation);
    m_offsetA.resize(nStation);
    m_offsetB.resize(nStation);
    m_offsetC.resize(nStation);

    for (unsigned int iB = 0; iB < m_idealGeometry.size(); ++iB)
    {
      double xBias = offset(offsetX);
      double yBias = offset(offsetY);
      double zBias = offset(offsetZ);
      double aBias = offset(offsetA);
      double bBias = offset(offsetB);
      double cBias = offset(offsetC);

      m_offsetX[iB] = xBias;
      m_offsetY[iB] = yBias;
      m_offsetZ[iB] = zBias;
      m_offsetA[iB] = aBias;
      m_offsetB[iB] = bBias;
      m_offsetC[iB] = cBias;
    }

    //
    // Then the set of toy tracks is produced
    //

    input_tracks->clear();

    my_tracks->PrepareStore(input_tracks,m_idealGeometry,m_params.nEvent());

    cout << "Number of tracks generated " << input_tracks->size() << endl; 

    Tracks::const_iterator itrack;

    //
    // Millepede tool is initialized
    //

    m_align->initialize(m_idealGeometry);

    for (itrack = input_tracks->begin(); itrack != input_tracks->end(); ++itrack) {
      
      // Smear because of detector resolution

      std::vector<double> smearX, smearY;    
      for (unsigned int iSmear = 0 ; iSmear <  nStation ; ++iSmear)
      {
        smearX.push_back(m_GaussGenerator->GetRandom());
        smearY.push_back(m_GaussGenerator->GetRandom());
      }            

      // Feed Millepede with the track (and the offsets)

      m_align->addTrack(*itrack, m_offsetX, m_offsetY, m_offsetZ,
			m_offsetA, m_offsetB, m_offsetC,
			smearX, smearY);
    }

    //
    // ...finally make the alignment and print the results
    //

    m_align->globalFit(); 

    const std::vector<double>& par    = m_align->parameters();
    const std::vector<double>& errors = m_align->errors(); 
    const std::vector<double>& pulls  = m_align->pulls(); 

    //
    // then we compute global offset and shearing (need to suppress them because 
    // internal alignment is not sensitive to them)
    //

    double z_moy =0;
    double sigma_z_moy = 0;

    // z_average  
    for (unsigned int j=0; j<m_idealGeometry.size(); j++) z_moy += m_idealGeometry[j].z();
    z_moy /= m_idealGeometry.size();

    // Error on z_moy
    for (unsigned int j=0; j<m_idealGeometry.size(); j++) sigma_z_moy += ( m_idealGeometry[j].z()-z_moy)*
							    ( m_idealGeometry[j].z()-z_moy);
    sigma_z_moy /= m_idealGeometry.size();

    double off_x = 0.;   // Global offsetl
    double shear_x = 0.; // Global shearing

    double off_y = 0.;  
    double shear_y = 0.;

    double off_z = 0.;  
    double scale_z = 0.;

    double off_a = 0.;   
    double off_b = 0.;   
    double off_c = 0.;   

    for (unsigned int iB = 0; iB < m_idealGeometry.size(); ++iB)
    {
      off_x += m_offsetX[iB];  
      shear_x += m_offsetX[iB]*(m_idealGeometry[iB].z()-z_moy)/sigma_z_moy;    

      off_y += m_offsetY[iB];  
      shear_y += m_offsetY[iB]*(m_idealGeometry[iB].z()-z_moy)/sigma_z_moy;

      off_z += m_offsetZ[iB];  
      scale_z += m_offsetZ[iB]*(m_idealGeometry[iB].z()-z_moy)/sigma_z_moy;

      off_a += m_offsetA[iB];  
      off_b += m_offsetB[iB];  
      off_c += m_offsetC[iB];  
    }

    off_x /= float(m_idealGeometry.size());
    shear_x /= float(m_idealGeometry.size()); 

    off_y /= float(m_idealGeometry.size());
    shear_y /= float(m_idealGeometry.size());  

    off_z /= float(m_idealGeometry.size());
    scale_z /= float(m_idealGeometry.size()); 

    off_a /= float(m_idealGeometry.size());
    off_b /= float(m_idealGeometry.size());
    off_c /= float(m_idealGeometry.size());

    int nPlane = m_idealGeometry.size();

    std::cout << "" << std::endl;
    std::cout << "--------------------------------------------------------" << std::endl;
    std::cout << "|                    FINAL RESULT                      |" << std::endl;
    std::cout << "--------------------------------------------------------" << std::endl;
    std::cout << "|  Gener.  |  Recon.  |   Diff.  |   Error  |    Cov.  |" << std::endl;
    std::cout << "--------------------------------------------------------" << std::endl;

    for (unsigned int iPlane = 0; iPlane <  m_offsetX.size() ; ++iPlane)
    {

      // What you measure with MILLEPEDE is the misalignment value CORRECTED 
      //from the offsets
      // you are not sensitive to the offsets
      //
      // Here we print the corrected values 
      //
            
      m_offsetX[iPlane] = m_offsetX[iPlane]-off_x-shear_x*(m_idealGeometry[iPlane].z()-z_moy);
      m_offsetY[iPlane] = m_offsetY[iPlane]-off_y-shear_y*(m_idealGeometry[iPlane].z()-z_moy);
      m_offsetZ[iPlane] = m_offsetZ[iPlane]-off_z-scale_z*(m_idealGeometry[iPlane].z()-z_moy);

      m_offsetA[iPlane] = m_offsetA[iPlane]-off_a;
      m_offsetB[iPlane] = m_offsetB[iPlane]-off_b;
      m_offsetC[iPlane] = m_offsetC[iPlane]-off_c;


      m_mis_bef[0]->Fill(m_offsetX[iPlane]);
      m_mis_bef[1]->Fill(m_offsetY[iPlane]);
      m_mis_bef[2]->Fill(m_offsetZ[iPlane]);
      m_mis_bef[3]->Fill(m_offsetA[iPlane]);
      m_mis_bef[4]->Fill(m_offsetB[iPlane]);
      m_mis_bef[5]->Fill(m_offsetC[iPlane]);

      m_mis_aft[0]->Fill(par[iPlane]          - m_offsetX[iPlane]);
      m_mis_aft[1]->Fill(par[iPlane+nPlane]   - m_offsetY[iPlane]);
      m_mis_aft[2]->Fill(par[iPlane+2*nPlane] - m_offsetZ[iPlane]);
      m_mis_aft[3]->Fill(par[iPlane+3*nPlane] - m_offsetA[iPlane]);
      m_mis_aft[4]->Fill(par[iPlane+4*nPlane] - m_offsetB[iPlane]);
      m_mis_aft[5]->Fill(par[iPlane+5*nPlane] - m_offsetC[iPlane]);

      m_pullHistos[0]->Fill(pulls[iPlane]);
      m_pullHistos[1]->Fill(pulls[iPlane+nPlane]);
      m_pullHistos[2]->Fill(pulls[iPlane+2*nPlane]);
      m_pullHistos[3]->Fill(pulls[iPlane+3*nPlane]);
      m_pullHistos[4]->Fill(pulls[iPlane+4*nPlane]);
      m_pullHistos[5]->Fill(pulls[iPlane+5*nPlane]);
      
      std::cout << "| " << std::setw(8) << m_offsetX[iPlane] 
		<< " | " << std::setw(8) << par[iPlane] 
		<< " | " << std::setw(8) << m_offsetX[iPlane] - par[iPlane] 
		<< " | " << std::setw(8) << errors[iPlane]
		<< " | " << std::setw(8) << sqrt(abs(errors[iPlane])) << " |" << std::endl;

      std::cout << "| " << std::setw(8) << m_offsetY[iPlane] 
		<< " | " << std::setw(8) << par[iPlane+nPlane] 
		<< " | " << std::setw(8) << m_offsetY[iPlane] - par[iPlane+nPlane] 
		<< " | " << std::setw(8) << errors[iPlane+nPlane]
		<< " | " << std::setw(8) << sqrt(abs(errors[iPlane+nPlane])) << " |" << std::endl;

      std::cout << "| " << std::setw(8) << m_offsetZ[iPlane] 
		<< " | " << std::setw(8) << par[iPlane+2*nPlane] 
		<< " | " << std::setw(8) << m_offsetZ[iPlane] - par[iPlane+2*nPlane] 
		<< " | " << std::setw(8) << errors[iPlane+2*nPlane]
		<< " | " << std::setw(8) << sqrt(abs(errors[iPlane+2*nPlane])) << " |" << std::endl;

      std::cout << "| " << std::setw(8) << m_offsetA[iPlane] 
		<< " | " << std::setw(8) << par[iPlane+3*nPlane] 
		<< " | " << std::setw(8) << m_offsetA[iPlane] - par[iPlane+3*nPlane] 
		<< " | " << std::setw(8) << errors[iPlane+3*nPlane]
		<< " | " << std::setw(8) << sqrt(abs(errors[iPlane+3*nPlane])) << " |" << std::endl;

      std::cout << "| " << std::setw(8) << m_offsetB[iPlane] 
		<< " | " << std::setw(8) << par[iPlane+4*nPlane] 
		<< " | " << std::setw(8) << m_offsetB[iPlane] - par[iPlane+4*nPlane] 
		<< " | " << std::setw(8) << errors[iPlane+4*nPlane]
		<< " | " << std::setw(8) << sqrt(abs(errors[iPlane+4*nPlane])) << " |" << std::endl;

      std::cout << "| " << std::setw(8) << m_offsetC[iPlane] 
		<< " | " << std::setw(8) << par[iPlane+5*nPlane] 
		<< " | " << std::setw(8) << m_offsetC[iPlane] - par[iPlane+5*nPlane] 
		<< " | " << std::setw(8) << errors[iPlane+5*nPlane]
		<< " | " << std::setw(8) << sqrt(abs(errors[iPlane+5*nPlane])) << " |" << std::endl;
           
      std::cout << "-------------------------------------------------------" << std::endl;

    } //iPlan
  }

  m_align->reset(); 

  return true;

}


/*
  FINALIZE:  
*/

bool myAlg::finalize()
{

  m_txHisto->Write();
  m_xyHisto->Write();
  m_rootFile->Write();	 
  std::cout << "*** Thats all folks ***" <<std::endl;	 
  return true;
}


